@@ -1,7 +1,7 @@
 #!/usr/bin/env python2.7
 # coin.py: a coin implemented on top of pybc
 
-import struct, hashlib, traceback, time, threading, logging, itertools
+import struct, hashlib, traceback, time, threading, logging, itertools, sys
 
 try:
     import pyelliptic
@@ -512,17 +512,40 @@ class CoinState(State):
         # Get the hash of the transaction
         transaction_hash = transaction.transaction_hash()
         
+        logging.debug("Applying transaction {}".format(util.bytes2string(
+            transaction_hash)))
+        
         for spent_output in transaction.inputs:
             # The inputs are in exactly the same tuple format as we use.
             # Remove each of them from the unused output set.
-            self.remove_unused_output(spent_output)
+            
+            if not self.has_unused_output(spent_output):
+                logging.error("Trying to remove nonexistent output: {} {} {} "
+                    "{}".format(util.bytes2string(spent_output[0]), 
+                    spent_output[1], spent_output[2], 
+                    util.bytes2string(spent_output[3])))
+                    
+                self.audit()
+                
+                # See if things just work out
+            else:
+                self.remove_unused_output(spent_output)
         
         for i, output in enumerate(transaction.outputs):
             # Unpack each output tuple
             amount, destination = output
             
+            # Make a tupe for the full unused output
+            unused_output = (transaction_hash, i, amount, destination)
             
-            self.add_unused_output((transaction_hash, i, amount, destination))
+            if self.has_unused_output(unused_output):
+                loging.error("Trying to create duplicate output: {} {} {} "
+                    "{}".format(util.bytes2string(unused_output[0]), 
+                    unused_output[1], unused_output[2], 
+                    util.bytes2string(unused_output[3])))
+                # See if things just work out
+            else:
+                self.add_unused_output(unused_output)
                 
     def remove_transaction(self, transaction):
         """
@@ -533,19 +556,46 @@ class CoinState(State):
         # Get the hash of the transaction
         transaction_hash = transaction.transaction_hash()
         
+        logging.debug("Reverting transaction {}".format(util.bytes2string(
+            transaction_hash)))
+        
         for spent_output in transaction.inputs:
             # The inputs are in exactly the same tuple format as we use. Add
             # them back to the unused output set, pointing to our useless value
             # of "".
-            self.add_unused_output(spent_output)
+            
+            if self.has_unused_output(spent_output):
+                logging.error("Trying to create duplicate output: {} {} {} "
+                    "{}".format(util.bytes2string(spent_output[0]), 
+                    spent_output[1], spent_output[2], 
+                    util.bytes2string(spent_output[3])))
+                
+                self.audit()
+                    
+                # See if things just work out
+            else:
+                self.add_unused_output(spent_output)
         
         for i, output in enumerate(transaction.outputs):
             # Unpack each output tuple
             amount, destination = output
             
-            # Remove its record from the set of unspent outputs
-            self.remove_unused_output((transaction_hash, i, amount, 
-                destination))
+            # Make a tupe for the full unused output
+            unused_output = (transaction_hash, i, amount, destination)
+            
+            if not self.has_unused_output(unused_output):
+                logging.error("Trying to remove nonexistent output: {} {} {} "
+                    "{}".format(util.bytes2string(unused_output[0]), 
+                    unused_output[1], unused_output[2], 
+                    util.bytes2string(unused_output[3])))
+                    
+                # Audit ourselves
+                self.audit()
+                
+                # See if things just work out
+            else:
+                # Remove its record from the set of unspent outputs
+                self.remove_unused_output(unused_output)
                     
     def step_forwards(self, block):
         """
@@ -561,9 +611,6 @@ class CoinState(State):
             # Parse the transaction
             transaction = Transaction.from_bytes(transaction_bytes)
         
-            # Get the hash of the transaction
-            transaction_hash = transaction.transaction_hash()
-            
             self.apply_transaction(transaction)
             
         if self.get_hash() != block.state_hash:
@@ -585,16 +632,16 @@ class CoinState(State):
             raise Exception("Stepping back block for state {} when actually in "
                 "state {}.".format(pybc.util.bytes2string(block.state_hash), 
                 pybc.util.bytes2string(self.get_hash())))
+        
+        logging.debug("Correctly in state {} before going back.".format(
+            pybc.util.bytes2string(self.get_hash())))
     
-        for transaction_bytes in transactions.unpack_transactions(
-            block.payload):
+        for transaction_bytes in reversed(list(transactions.unpack_transactions(
+            block.payload))):
             
             # Parse the transaction
             transaction = Transaction.from_bytes(transaction_bytes)
             
-            # Get the hash of the transaction
-            transaction_hash = transaction.transaction_hash()
-            
             self.remove_transaction(transaction)
             
     def get_unused_outputs(self):
@@ -725,6 +772,15 @@ class CoinState(State):
         """
         
         return StateMachine(AuthenticatedDictionaryStateComponent, self)
+        
+    def audit(self):
+        """
+        Make sure the State is internally consistent.
+        
+        """
+        
+        # Audit our AuthenticatedDictionary
+        self.unused_outputs.audit()
     
     def commit(self):
         """
@@ -976,6 +1032,9 @@ class CoinBlockchain(TransactionalBlockchain):
             # other, so we can safely advance the State with them.
             next_state = self.state.copy()
             
+            # How many bytes of transaction data have we used?
+            transaction_bytes_used = 0
+            
             for transaction_bytes in self.transactions.values():
                 # Parse this transaction out.
                 transaction = Transaction.from_bytes(transaction_bytes)
@@ -993,6 +1052,12 @@ class CoinBlockchain(TransactionalBlockchain):
                         # it earlier, but we should probably check it again.
                         to_include.append(transaction)
                         total_fee += fee
+                        transaction_bytes_used += len(transaction_bytes)
+                        
+                if transaction_bytes_used >= 1024 * 1024:
+                    # Don't make blocks bigger than 1 MB of transaction data
+                    logging.info("Hit block size limit for generation.")
+                    break
             
             # Add a transaction that gives all the generated coins and fees to
             # us.
